#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

class ValuesFromDictFonctor:
    """
    The :py:class:`ValuesFromDictFonctor` class is an
    internal minifold functor used to extract a subset of values
    from a dictionary.

    It is used by several minifold :py:class:`Connector` classes,
    including:

    - :py:class:`GroupByConnector`;
    - :py:class:`SortByConnector`;
    - :py:class:`UniqueConnector`.
    """
    def __init__(self, attributes: list):
        """
        Constructor.

        Args:
            attributes (list): The keys of the values to be extracted.
        """
        self.m_attributes = attributes

    @property
    def attributes(self) -> list:
        """
        Retrieves the keys of interest related to this
        :py:class:`ValuesFromDictFonctor` instance.

        Returns:
            The attributes of interest related to this
            :py:class:`ValuesFromDictFonctor` instance.
        """
        return self.m_attributes

    def __call__(self, entry: dict) -> tuple:
        """
        Applies this py:class:`ValuesFromDictFonctor` instance
        to an input dictionary.

        Args:
            entry (dict): An input dictionary.

        Returns:
            The extracted values of interest.
        """
        return tuple(entry.get(k) for k in self.attributes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import json, os, pickle
from functools import partial
from .cache import DEFAULT_CACHE_STORAGE_BASE_DIR
from .entries_connector import EntriesConnector
from .filesystem import mkdir
from .cache import JsonCacheConnector
from .log import Log

class CachedEntriesConnector(EntriesConnector):
    """
    :py:class:`CachedEntriesConnector` is a :py:class:`Connector` is an
    abstract class used to fetch data from a cache saved on the local storage.

    See specializations:

    - :py:class:`JsonCachedConnector` (caching using a JSON file)
    - :py:class:`PickleCachedConnector` (caching using pickles)
    """
    def __init__(
        self,
        load_entries: callable,
        cache_filename: str,
        load_cache: callable,
        save_cache: callable,
        read_mode: str,
        write_mode: str,
        with_cache: bool = True
    ):
        """
        Constructor.

        Args:
            load_entries (callable): A function called to populate this
                :py:class:`CachedEntriesConnector`.
            cache_filename (str): The path to the file used to save the
                cache on the local storage.
            load_cache (callable): A function `entries = load_cache(f)` where:
                `entries` is a list of dictionnaries;
                `f` is the (read) file descriptor of the cache.
            save_cache (callable): A function `save_cache(cache_filename, f)` where:
                `entries` is a list of dictionnaries;
                `f` is the (write) file descriptor of the cache.
            read_mode (str): A string specifying how the read file decriptor of the
                cache must be created. Possible values are ``"r"`` (text-based cache)
                and ``"rb"`` (binary cache).
            write_mode (str): A string specifying how the write file decriptor of the
                cache must be created. Possible values are ``"w"`` (text-based cache)
                and ``"wb"`` (binary cache).
        """
        loaded_from_cache = False
        if with_cache:
            try:
                with open(cache_filename, read_mode) as f:
                    Log.info("%s: Loading cache from [%s]" % (type(self), cache_filename))
                    entries = load_cache(f)
                    Log.info("Loaded %d entries" % len(entries))
                    loaded_from_cache = True
            except FileNotFoundError:
                Log.debug("%s: Cache [%s] not found" % (type(self), cache_filename))
                pass
            except Exception as e:
                Log.debug("%s: Cache [%s] corrupted" % (type(self), cache_filename))
                Log.error(e)
                pass

        # Parse the input data (if needed)
        if not loaded_from_cache:
            entries = load_entries()
            Log.info("Loaded %d entries" % len(entries))

        # Save into cache (if needed)
        if with_cache and not loaded_from_cache:
            Log.info("%s: Saving data into cache [%s]" % (type(self), cache_filename))
            mkdir(os.path.dirname(cache_filename))
            with open(cache_filename, write_mode) as f:
                save_cache(entries, f)

        super().__init__(entries)


class JsonCachedConnector(CachedEntriesConnector):
    """
    The :py:class:`JsonCachedConnector` class implements the
    :py:class:`CachedEntriesConnector` using a tierce JSON file.
    """
    def __init__(self, load_entries, cache_filename :str, **kwargs):
        """
        Constructor.
        See also :py:class:`CachedEntriesConnector.__init__`

        Args:
            load_entries (callable): A function called to populate this
                :py:class:`CachedEntriesConnector`.
            cache_filename (str): The path to the file used to save the
                cache on the local storage.
        """
        super().__init__(
            load_entries,
            cache_filename,
            load_cache = json.load,
            save_cache = partial(json.dump, indent=4),
            read_mode  = "r",
            write_mode = "w",
            **kwargs
        )


class PickleCachedConnector(CachedEntriesConnector):
    """
    The :py:class:`PickleCachedConnector` class implements the
    :py:class:`CachedEntriesConnector` using a tierce pickle file.
    """
    def __init__(self, load_entries, cache_filename :str, **kwargs):
        """
        Constructor.
        See also :py:class:`CachedEntriesConnector.__init__`

        Args:
            load_entries (callable): A function called to populate this
                :py:class:`CachedEntriesConnector`.
            cache_filename (str): The path to the file used to save the
                cache on the local storage.
        """
        super().__init__(
            load_entries,
            cache_filename,
            load_cache = pickle.load,
            save_cache = pickle.dump,
            read_mode  = "rb",
            write_mode = "wb",
            **kwargs
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['minifold']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4',
 'ldap3',
 'lxml',
 'pycountry',
 'pymongo',
 'requests',
 'requests-cache',
 'tweepy',
 'urllib3',
 'xmltodict']

setup_kwargs = {
    'name': 'minifold',
    'version': '0.10.0',
    'description': 'Minifold is a Python module able to interact with various data sources (e.g. CSV, LDAP, SQL, twitter, etc.) and to query/combine/aggregate them with database-like operators.',
    'long_description': '# Minifold\n\n[![PyPI](https://img.shields.io/pypi/v/minifold.svg)](https://pypi.python.org/pypi/minifold/)\n[![Build](https://github.com/nokia/minifold/workflows/build/badge.svg)](https://github.com/nokia/minifold/actions/workflows/build.yml)\n[![Documentation](https://github.com/nokia/minifold/workflows/docs/badge.svg)](https://github.com/nokia/minifold/actions/workflows/docs.yml)\n[![ReadTheDocs](https://readthedocs.org/projects/minifold/badge/?version=latest)](https://minifold.readthedocs.io/en/)\n[![codecov](https://codecov.io/gh/nokia/minifold/branch/master/graph/badge.svg?token=OZM4J0Y2VL)](https://codecov.io/gh/nokia/minifold)\n\n## Overview\n\n[Minifold](https://github.com/nokia/minifold.git) is a [Python](http://python.org/) module able to interact with various data sources (e.g. CSV, LDAP, SQL, twitter, etc.) and to query/combine/aggregate them with database-like operators.\n\n## Use cases\n\nThis framework has been in various Nokia projects. It also used in the [LINCS](https://www.lincs.fr) website to generate:\n\n* [trombinoscope](https://www.lincs.fr/people/)\n* [homepages](https://www.lincs.fr/people/?more=marc_olivier_buob)\n* [co-author graph](https://www.lincs.fr/research/lincs-graph/), thanks to [pyBGL](https://github.com/nokia/pybgl.git) and [GraphViz](http://graphviz.org/)\n\nFor more information, feel free to visit the [wiki](https://github.com/nokia/minifold/wiki):\n\n* [Overview](https://github.com/nokia/minifold/wiki/Overview)\n* [Installation](https://github.com/nokia/minifold/wiki/Installation)\n* [Tutorial](https://github.com/nokia/minifold/wiki/Tutorial)\n* [Design](https://github.com/nokia/minifold/wiki/Design)\n* [Configuration](https://github.com/nokia/minifold/wiki/Configuration)\n* [Framework](https://github.com/nokia/minifold/wiki/Framework)\n* [Tests](https://github.com/nokia/minifold/wiki/Tests)\n* [Packaging](https://github.com/nokia/minifold/wiki/Packaging)\n\n# License\n\nThis project is licensed under the BSD-3-Clause license - see the [LICENSE](https://github.com/nokia/minifold/blob/master/LICENSE).\n',
    'author': 'Marc-Olivier Buob',
    'author_email': 'marc-olivier.buob@nokia-bell-labs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

import codecs, json
from .entries_connector import EntriesConnector

def identity(x: object) -> list:
    """
    Identity function.

    Args:
        x (object): An arbitrary object.

    Returns:
        The ``x`` object.
    """
    return x

def load_json_from_str(json_data: str, extract_json: callable = identity) -> list:
    """
    Loads minifold from a JSON string.

    Args:
        json_data (str): The JSON data.
        extract_json (callable): A function that reprocesses the python structure
            resulting from the JSON to produce a list of minifold entries.
            Defaults to :py:func:`identity`.

    Returns:
        A list of minifold entries.

    """
    assert isinstance(json_data, str)
    entries = extract_json(json.loads(json_data))
    assert isinstance(entries, list), f"Minifold expects a list of dict, got {type(entries)}"
    return entries

def load_json_from_file(json_filename: str, extract_json: callable = identity) -> list:
    """
    Loads minifold from a JSON file.

    Args:
        json_filename (str): The path to a JSON file.
        extract_json (callable): A function that reprocesses the python structure
            resulting from the JSON to produce a list of minifold entries.
            Defaults to :py:func:`identity`.

    Returns:
        A list of minifold entries.
    """
    with codecs.open(json_filename, encoding = "utf8") as f_json:
        entries = load_json_from_str(f_json.read(), extract_json)
    return entries

class JsonConnector(EntriesConnector):
    """
    The :py:class:`JsonConnector` class is a gateway to a JSON string.
    """
    def __init__(self, json_data: str, extract_json: callable = identity):
        """
        Constructor.

        Args:
            json_data: A JSON string.
            extract_json: A function that converts the JSON data to a list
                of minifold entries. Defaults to :py:func:`identity`.
        """
        entries = load_json_from_str(json_data, extract_json)
        assert isinstance(entries, list)
        super().__init__(entries)

class JsonFileConnector(JsonConnector):
    """
    The :py:class:`JsonConnector` class is a gateway to a JSON file.
    """
    def __init__(self, json_filename: str, extract_json: callable = identity):
        """
        Constructor.

        Args:
            json_filename: The path of the input JSON file.
            extract_json: A function that converts the JSON data to a list
                of minifold entries. Defaults to :py:func:`identity`.
        """
        with codecs.open(json_filename, encoding = "utf8") as f_json:
            super().__init__(f_json.read(), extract_json)

    # TODO: For the moment we only support ACTION_READ queries,
    # but this could be extended



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

"""
This file integrates the classes provided in ``scholar.py`` to minifold.
It allows to query `Google scholar <https://scholar.google.com/>`__.
"""

import sys

try:
    from bs4 import BeautifulSoup
except ImportError as e:
    print(
        "\n  ".join([
            "Please install bs4",
            "APT: sudo apt install python3-bs4",
            "PIP: sudo pip3 install bs4",
        ]),
        file=sys.stderr
    )
    raise e

import difflib, operator
from .binary_predicate import BinaryPredicate
from .connector import Connector
from .download import download
from .log import Log
from .query import ACTION_READ, Query
from .scholar import *


def parse_article(s_html: str) -> dict:
    """
    Parse a "gs_res_ccl_mid" div (wrapping each article) returned by Google Scholar.

    Args:
        s_html (str): The HTML string containing the div.

    Returns:
        The dict describing the article, structured as follows:

        - "authors"       : list, The list only contains the first authors. The last author name maybe incomplete.
        - "cluster_id"    : int,
        - "conference"    : str,  # May be unset
        - "editor"        : str,
        - "excerpt"       : str,  # May be unset
        - "num_citations" : int,
        - "num_versions"  : int,
        - "title"         : str,
        - "url_citations" : str,
        - "url_versions"  : str,
        - "url_pdf"       : str,
        - "url_title"     : str,
        - "year"          : int,

        An incomplete string may start by ``'…'`` and ends with ``'…'``.
        URLs are absolute.
    """

    def clean_string(s: str) -> str:
        """
        Cleans a string (title, author, etc.) returned by Google Scholar.

        Args:
            s (str): The input string.

        Returns:
            The cleaned string.
        """
        return re.sub("( |\\xa0|\\n)+", " ", s).strip()

    def extract_first_int(s: str) -> int:
        """
        Extracts the first int involved in a string.

        Args:
            s (str): The input string.

        Raises:
            ``AttributeError`` if ``s`` does not contain at least
            one integer.
        """
        return int(re.search("\\d+", s).group())

    entry = dict()
    soup = BeautifulSoup(s_html, features="lxml")

    # Extract url_title
    div = soup.find(name="div", attrs={"class": "gs_or_ggsm"})
    if div:
        a = div.find(name="a")
        if a:
            entry["url_pdf"] = a["href"]

    # Extract url_title, title
    h3 = soup.find(name="h3", attrs={"class": "gs_rt"})
    if h3:
        span = h3.find(name="span", attrs={"class": "gt_ct1"})
        if span:  # [CITATION]
            entry["doc_type"] = span.text[1:-1]
        a = h3.find(name="a")
        if a:
            entry["url_title"] = a["href"]
            entry["title"] = clean_string(a.text)

    # Extract authors, year, editor
    div = soup.find(name="div", attrs={"class": "gs_a"})
    if div:
        s = div.text.replace("\xa0", " ")
        (authors, conference_year, editor) = s.split(" - ")
        entry["authors"] = [clean_string(a) for a in authors.split(",")]
        conference_year = [clean_string(elt) for elt in conference_year.split(",")]
        entry["conference"] = conference_year[0] if len(conference_year) == 2 else None
        entry["year"] = int(conference_year[-1])
        entry["editor"] = clean_string(editor)

    # Extract exceprt
    div = soup.find(name="div", attrs={"class": "gs_rs"})
    if div:
        entry["excerpt"] = clean_string(div.text.strip())

    # Extract num_citations and num_versions
    divs = soup.findAll(name="div", attrs={"class": "gs_fl"})
    if divs and len(divs) > 1:
        links = divs[1].findAll("a")
        if links:
            # It is important to not use link index here as it may change depending on the
            # document (e.g whether it is cited or not).
            for link in links:
                s = link.text.lower()
                key = (
                    "citations" if "cited by" in s else
                    "versions"  if "versions" in s else
                    None
                )
                if key:
                    entry["num_%s" % key] = extract_first_int(s)
                    entry["url_%s" % key] = ScholarConf.SCHOLAR_SITE + link["href"]
                    entry["cluster_id"] = extract_first_int(link["href"])
    return entry


class MinifoldScholarQuerier(ScholarQuerier):
    """
    :py:class`MinifoldScholarQuerier` overloads :py:class:`ScholarQuerier`
    to fetch more attributes.
    """
    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        self.articles = list()

    def parse(self, s_html: str):
        """
        Populates :py:attr:`self.articles` using the HTML Google Scholar page.

        Args:
            s_html (str): An HTML Google Scholar page content.
        """
        soup = SoupKitchen.make_soup(s_html)
        soup = soup.find(name="div", attrs={"id": "gs_res_ccl_mid"})
        for div in soup.findAll(name="div", attrs={"class": "gs_r"}):
            s = div.prettify()
            entry = parse_article(s)
            self.articles.append(entry)

    def send_query(self, gs_query: ScholarQuery):
        """
        Sends a query to Google scholar.

        Args:
            gs_query (ScholarQuery): A Google scholar query.

        Raises:
            ``RuntimeError`` if the result can't be fetched from Google scholar.
        """
        # Network
        url = gs_query.get_url()
        Log.info("GoogleScholar <-- %s" % url)

        # Here, we rely on minifold downloader as it can cache results.
        # s_html = self._get_http_response(url)
        response = download(url)
        s_html = response.text

        # Parsing
        if s_html:
            self.parse(s_html)
            self.articles = list()
            self.parse(s_html)
        else:
            raise RuntimeError("Cannot fetch result from Google scholar")


class GoogleScholarConnector(Connector):
    """
    The :py:class:`GoogleScholarConnector` class is a minifold gateway
    allowing to query `Google scholar <https://scholar.google.com/>`.

    See also:
    - :py:class:`DblpConnector`.
    - :py:class:`GoogleScholarConnector`.

    """
    def __init__(self, citation_format: str = ScholarSettings.CITFORM_BIBTEX):
        """
        Constructor.

        Args:
            citation_format (str): The citation format.
        """
        settings = ScholarSettings()
        settings.set_citation_format(citation_format)
        self.querier = MinifoldScholarQuerier()  # ScholarQuerier()
        self.querier.apply_settings(settings)

    def attributes(self, object: str) -> set:
        """
        Lists available attributes related to a given collection of object
        stored in this :py:class:`GoogleScholarConnector` instance.

        Args:
            object (str): The name of the collection of entries.
                The only supported object is: ``"publication"`` and ``"cluster"``.

        Returns:
            The set of available attributes for ``object``.
        """
        if object == "publication" or not object:
            return {
                "authors",  # str
                "cluster_id",  # int,
                "conference",  # str,
                "doc_type",  # str,
                "editor",  # str,
                "excerpt",  # str,
                "num_citations",  # int,
                "num_versions",  # int,
                "title",  # str,
                "url_citations",  # str,
                "url_versions",  # str,
                "url_pdf",  # str,
                "url_title",  # str,
                "year",  # int,
            }
        elif object == "cluster":
            raise RuntimeError("Object %s is not yet supported" % object)
        else:
            raise RuntimeError("Invalid object %s" % object)

    @staticmethod
    def filter_to_scholar(p: BinaryPredicate, gs_query: SearchScholarQuery, authors: list):
        """
        Converts a minifold predicate (applying to authors) to the corresponding
        scholar filter (recursive function).

        Args:
            p (BinaryPredicate): The minifold filter.
            gs_query (ScholarQuery): The Google Scholar query.
            authors (list): Pass an empty list.
        """
        if p.operator == operator.__and__:
            GoogleScholarConnector.filter_to_scholar(p.left, gs_query, authors)
            GoogleScholarConnector.filter_to_scholar(p.right, gs_query, authors)
            return

        if isinstance(p.left, BinaryPredicate) or isinstance(p.right, BinaryPredicate):
            raise RuntimeError("Invalid clause %s" % p)

        attr = p.left
        value = p.right
        if not isinstance(attr, str):
            raise RuntimeError("Invalid left operand %s" % p)

        if attr == "authors":
            if p.operator in {operator.__eq__, operator.__contains__}:
                authors.append(value)
                gs_query.set_author(value)
            else:
                raise RuntimeError("Invalid operator %s" % p)
        #       elif attr == "words":
        #           if p.operator == operator.__eq__ or p.operator == operator.__ge__:
        #               gs_query.set_words(options.allw)       # All of these words must appear
        #           elif p.operator == operator.__le__:
        #               gs_query.set_words_some(options.some)  # Some of these words must appear
        #           elif p.operator == operator.__ne__:
        #               gs_query.set_words_none(options.none)  # None of these words must appear
        #           else:
        #               raise RuntimeError("Invalid operator %s" % p)
        elif attr == "phrase":
            Log.warning("GoogleScholarConnector: filtering on %s not yet implemented" % attr)
            # gs_query.set_phrase(options.phrase)
            pass
        elif attr == "doc_type":
            Log.warning("GoogleScholarConnector: filtering on %s not yet implemented" % attr)
            # NOT YET IMPLEMENTED
            # gs_query.set_include_patents(False)
            # gs_query.set_include_citations(False)
            pass
        elif attr == "year":
            assert isinstance(value, int)
            (start, end) = gs_query.timeframe
            if p.operator == operator.__eq__:
                start = max(start, value) if start else value
                end = min(end, value) if end else value
            elif p.operator == operator.__le__:
                end = min(end, value) if end else value
            elif p.operator == operator.__lt__:
                end = min(end, value - 1) if end else value
            elif p.operator == operator.__ge__:
                start = max(start, value) if start else value
            elif p.operator == operator.__gt__:
                start = max(start, value + 1) if start else value
            else:
                raise RuntimeError("Invalid operator %s" % p)
            if start is not None and end is not None and end < start:
                raise RuntimeError("Invalid range of date")
            gs_query.set_timeframe(start, end)
        elif attr == "publication":
            if p.operator == operator.__eq__:
                gs_query.set_pub(value)
            else:
                raise RuntimeError("Invalid operator %s" % p)

    @staticmethod
    def sanitize_author(authors: list, author: str) -> str:
        """
        Fixes author names, by finding in an input list of strings the
        closest string with the input string.

        Args:
            authors (list): The list of strings.
            author (str): The reference string.

        Returns:
            The string of `authors` closest to `author` if any, else `author`
        """
        # Find the best match in authors. Indeed if we search "John Doe", Google Scholar
        # returns "J Doe" and so the self.reshape_entries will drop the records.
        # In practice, the cutoff must be quite low, but not too low. Otherwise,
        # co-authors may be wrongly renamed! Empirically, 0.4 is a good tradeoff).
        l = difflib.get_close_matches(author, authors, n=1, cutoff=0.4)
        if len(l) != 1:
            # We are trying to rename a co-author without knowing is real name.
            return author
        else:
            # We are renaming an author with his/her real name.
            return l[0]

    def query(self, query: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input Query.
        """

        super().query(query)
        if query.action != ACTION_READ:
            raise RuntimeError("Action not supported" % query.action)
        if query.filters and not isinstance(query.filters, BinaryPredicate):
            raise RuntimeError("Invalid filter: %s" % query.filters)

        authors = list()
        if query.object == "publication" or not query.object:
            gs_query = SearchScholarQuery()
            GoogleScholarConnector.filter_to_scholar(query.filters, gs_query, authors)
        # elif query.object == "cluster":
        #     gs_query = ClusterScholarQuery(cluster=options.cluster_id)
        else:
            raise RuntimeError("Invalid object %r" % query.object)

        # Craft the query
        if query.limit is not None:
            gs_query.set_num_page_results(min(query.limit, ScholarConf.MAX_PAGE_RESULTS))

        # Doc type.
        gs_query.set_include_citations(False)
        gs_query.set_include_patents(True)

        # Send the query
        self.querier.send_query(gs_query)

        # Extract results
        if isinstance(self.querier, MinifoldScholarQuerier):
            entries = self.querier.articles
        else:
            entries = [
                {k: v[0] for (k, v) in article.attrs.items()}
                for article in self.querier.articles
            ]

        if gs_query.author:
            # Sanitize author list.
            for entry in entries:
                entry["authors"] = [
                    GoogleScholarConnector.sanitize_author(authors, author)
                    for author in entry["authors"]
                ]
        entries = self.reshape_entries(query, entries)
        return self.answer(query, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from .connector import Connector
from .query import Query, ACTION_READ

def count_gen(gen: iter) -> int:
    """
    Count the number of elements listed by a generator.

    Args:
        gen (iter): A generator.

    Returns:
        The number of elements listed by a generator.
    """
    n = 0
    for _ in gen:
        n += 1
    return n

def count(entries: list) -> int:
    """
    Count the number of entries.

    Example:
        >>> count([{"a": 10}, {"a": 20}])
        2

    Args:
        entries (list): A list of entries.

    Returns:
        The number of element in entries.
    """
    return len(entries)

class CountConnector(Connector):
    """
    The :py:class:`CountConnector` class is used to implement the COUNT
    statement in a minifold query plan. As it is one of the rare
    connector returning an integer (instead of a list of entries) this
    is often the root connector in the tree modeling the minifold query plan.
    """
    def __init__(self, child :Connector):
        """
        Constructor.

        Args:
            child (Connector): The child :py:class:`Connector` instance.
        """
        super().__init__()
        self.m_child = child

    def query(self, query: Query) -> int:
        """
        Handles a minifold query.

        Args:
            query (Query): The handled :py:class:`Query` instance.

        Raises:
            ValueError: if ``query.action != ACTION_READ``.

        Returns:
            The number of entries matched by this :py:class:`Query`.
        """
        super().query(query)
        if query.action != ACTION_READ:
            raise ValueError("Invalid action in query %s" % query)
        return count(self.m_child.query(query))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

try:
    from ldap3                  import Server, Connection, SUBTREE, ALL_ATTRIBUTES, ALL
    from ldap3.core.exceptions  import LDAPInvalidFilterError
except ImportError as e:
    from .log import Log
    Log.warning(
        "Please install ldap3.\n"
        "  APT: sudo apt install python3-ldap3\n"
        "  PIP: sudo pip3 install --upgrade ldap3\n"
    )
    raise e

import operator
from .connector import Connector
from .query import Query, ACTION_READ
from .binary_predicate import BinaryPredicate
from .log import Log

class LdapConnector(Connector):
    def __init__(
        self,
        ldap_host: str,
        ldap_user: str = None,
        ldap_password: str = None,
        ldap_use_ssl: bool = None
    ):
        """
        Constructor.

        Args:
            ldap_host (str): The FQDN or the IP of the server (e.g., "my-ldap.firm.com").
            ldap_user (str): The LDAP login used to connect to the server.
            ldap_password (str): The LDAP password of ``ldap_user``.
            ldap_use_ssl (bool): Pass ``True`` if the connection to the server must
                be established using SSL, ``False`` or ``None`` otherwise.
        """
        super().__init__()
        self.m_server     = Server(ldap_host, use_ssl=ldap_use_ssl, get_info=ALL)
        self.m_connection = Connection(self.m_server, ldap_user, ldap_password)
        self.m_connection.bind()

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`LdapConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return {str(key) for key in self.m_connection.server.schema.attribute_types.keys()}

    def __enter__(self):
        """
        Method called when entering a ``with LdapConnector(...):`` block.
        """
        self.m_connection.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        """
        Method called when leaving a ``with LdapConnector(...):`` block.
        """
        self.m_connection.unbind()

    @staticmethod
    def operator_to_ldap(op) -> str:
        """
        Converts a minifold operator to the corresponding LDAP operator.

        Args:
            operator: The operator to be converted.

        Returns:
            The corresponding LDAP string.
        """
        if op == operator.__eq__:
            return "="
        elif op == operator.__and__:
            return "&"
        elif op == operator.__or__:
            return "|"
        # elif op == "<=" or op == ">=" or op == "<" or op == ">" or op == "~": return op
        else:
            raise RuntimeError("LdapConnector::operator_to_ldap: op = %s not supported" % op)

    @staticmethod
    def binary_predicate_to_ldap(p: BinaryPredicate) -> str:
        """
        Converts a :py:class:`BinaryPredicate` to the corresponding LDAP predicate.

        Args:
            p (BinaryPredicate): A :py:class:`BinaryPredicate` instance.

        Returns:
            The corresponding LDAP predicate.
        """

        if p.operator in ["&&", "||"]:
            # See http://ldap3.readthedocs.io/searches.html
            fmt = "(%(operator)s%(left)s%(right)s)"
        else:
            fmt = "%(left)s%(operator)s%(right)s"
        return fmt % {
            "left": LdapConnector.operand_to_ldap(p.left),
            "operator": LdapConnector.operator_to_ldap(p.operator),
            "right": LdapConnector.operand_to_ldap(p.right)
        }

    @staticmethod
    def operand_to_ldap(operand: str) -> str:
        """
        Converts a minifold operand to the corresponding LDAP operand.

        Args:
            operand (str): The operand to be converted.

        Returns:
            The corresponding string.
        """
        if isinstance(operand, BinaryPredicate):
            return "(%s)" % LdapConnector.binary_predicate_to_ldap(operand)
        elif isinstance(operand, (int, float, str)):
            # Do not add '...' or "..." for str
            return operand
        else:
            raise RuntimeError("Operand type not supported %s" % type(operand))

    @staticmethod
    def literal_from_ldap(b: bytes) -> str:
        """
        Converts a LDAP literal to the corresponding string.

        Args:
            b (bytes): The input literal.

        Returns:
            The corresponding string.
        """
        try:
            return int(b)
        except ValueError:
            pass
        try:
            return float(b)
        except ValueError:
            pass
        try:
            return b.decode("utf-8")
        except ValueError:
            pass
        return b

    @staticmethod
    def sanitize_dict(d: dict) -> dict:
        """
        Reshapes the dictionary returned by a LDAP query to a minifold entry.

        Args:
            d (dict): The input dict.

        Returns:
            The corresponding sanitized dictionary.
        """
        sane = False
        for k,v in d.items():
            # raw values are always list
            if len(v) == 1:
                d[k] = LdapConnector.literal_from_ldap(v[0])
            else:
                d[k] = [LdapConnector.literal_from_ldap(x) for x in v]
            if not d[k]:
                d[k] = None
            else:
                sane = True
        return d if sane else dict()

    def query(self, q: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """
        super().query(q)
        entries = list()

        if q.action == ACTION_READ:
            if len(q.attributes) == 0:
                attributes = ALL_ATTRIBUTES
            else:
                attributes = q.attributes

            if q.filters is None:
                keep_if = "(objectClass=*)"
            else:
                keep_if = LdapConnector.operand_to_ldap(q.filters)

            try:
                if attributes != ALL_ATTRIBUTES:
                    attributes = set(attributes) & self.attributes(q.object)
                Log.info("--> LDAP: dn = %s filter = %s attributes = %s"
                    % (
                        q.object,
                        keep_if,
                        attributes
                    )
                )
                self.m_connection.search(
                    q.object,
                    keep_if,
                    search_scope = SUBTREE,
                    attributes = attributes
                )
            except LDAPInvalidFilterError as e:
                Log.error("LdapConnector::query: Invalid filter: %s" % keep_if)
                raise e

            for entry in self.m_connection.response:
                entry = LdapConnector.sanitize_dict(entry["raw_attributes"])
                if len(entry) > 0:
                    # Fix missing keys
                    if len(q.attributes) > 0:
                        for missing_key in (set(q.attributes) - set(entry.keys())):
                            entry[missing_key] = None
                    entries.append(dict(entry))
        else:
            raise RuntimeError("Not implemented")
        return self.answer(q, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from copy import deepcopy
from .connector import Connector
from .query import Query

def find_lambda_dependencies(func: callable) -> set:
    """
    Infers the keys needed by a function processing a dictionary
    to not trigger :py:class:`KeyError` exception.

    Args:
        func (callable): A function taking a dictionary in parameter.

    Returns:
        The keys needed by ``func`` to process a dictionary.
    """
    entry = dict()
    while True:
        try:
            _ = func(entry)
        except KeyError as e:
            needed_key = e.args[0]
            entry[needed_key] = None
            continue
        except Exception:
            pass
        break
    return set(entry.keys())

def find_lambdas_dependencies(map_lambdas: dict) -> dict:
    """
    Infers the keys needed by several functions that outputs a specific
    entry key when processing an input entry so that no
    :py:class:`KeyError` exception is raised.

    Args:
        map_lambdas (dict): A dictionary that maps key
            (existing or new) key attributes with a function
            processing an input entry.

    Returns:
        The keys needed by the functions involved in ``map_lambdas.values()``
        to process a minifold entry
    """
    return {
        attr: find_lambda_dependencies(func)
        for (attr, func) in map_lambdas.items()
    }

def lambdas(map_lambdas: dict, entries: list, attributes: set = None) -> list:
    """
    Be sure that the result is deterministic without regards each
    lambda is processed.

    Examples:
        >>> map_lambdas = {"x": lambda e: 10 + e["x"]} # OK
        >>> map_lambdas = {"x": lambda e: 10 + e["x"] + e["y"]} # OK
        >>> map_lambdas = {
        ...     "x": lambda e: 10 + e["x"] + e["y"],
        ...     "y": lambda e: 10 + e["y"]
        ... } # not OK because e["y"] is ambiguous.
    """
    attrs = set(map_lambdas.keys())
    if attributes:
        attrs &= set(attributes)
    for entry in entries:
        for attr, func in map_lambdas.items():
            if attr in attrs:
                try:
                    entry[attr] = func(entry)
                except KeyError:
                    entry[attr] = None
    return entries

class LambdasConnector(Connector):
    """
    The :py:class:`LambdasConnector` class is used to apply the `lambdas` function
    in the middle of a minifold pipeline. It allows to craft or reshape a flow
    of minifold entries on-the-fly.
    """
    def __init__(self, map_lambdas: dict, child: Connector, map_dependencies: dict = None):
        """
        Constructor.

        Args:
            map_lambdas (dict): A dictionary that maps key
                (existing or new) key attributes with a function
                processing an input entry.
            child (Connector): The child minifold :py:class:`Connector`
                instance.
        """
        # Ensure there is no cyclic dependency (see lambdas())
        super().__init__()
        self.m_child = child
        self.m_map_lambdas = map_lambdas

        # If no attribute dependency is provided, use find_lambda_dependencies heuristic.
        self.m_map_dependencies = map_dependencies \
            if map_dependencies is not None \
            else find_lambdas_dependencies(map_lambdas)

    @property
    def child(self) -> Connector:
        """
        Accessor to the child minifold :py:class:`Connector` instance.

        Returns:
            The child minifold :py:class:`Connector` instance.
        """
        return self.m_child

    def attributes(self, object: str) -> set:
        """
        Lists the available attributes related to a given collection of
        minifold entries exposed by this :py:class:`LambdasConnector` instance.

        Args:
            object (str): The name of the collection.

        Returns:
            The set of corresponding attributes.
        """
        return self.m_child.attributes(object) | set(self.m_map_lambdas.keys())

    def find_needed_attributes(self, attributes: set) -> set:
        """
        Finds the attribute that are needed to process a set of attributes of interest.

        Args:
            attributes (set): The attributes of interest.

        Returns:
            The corresponding needed attributes.
        """
        ret = set()
        for attribute in attributes:
            needed_attributes = self.m_map_dependencies.get(attribute, {attribute})
            ret |= needed_attributes
        return ret

    def query(self, q: Query) -> list:
        """
        Handles an input :py:class:`Query` instance.

        Args:
            query (Query): The handled query.

        Returns:
            The list of entries matching the input query.
        """

        super().query(q)
        q_child = deepcopy(q)

        if q.attributes or q.filters:
            where_attributes = self.find_needed_attributes(
                find_lambda_dependencies(q.filters)
            )

            if q.attributes:
                child_attributes = self.m_child.attributes(None)
                select_attributes  = self.find_needed_attributes(set(q.attributes))
                # Remove lambda-specific attributes
                lambdas_attributes = set(self.m_map_lambdas.keys()) - {
                    dep_attr
                    for attr in q.attributes
                    for dep_attr in self.m_map_dependencies.get(attr, set())
                    if dep_attr in child_attributes
                }
                q_child.attributes = list(
                    (select_attributes | where_attributes) - lambdas_attributes
                )

            if where_attributes:
                # If the filter involves at least one attribute of the LambdasConnector
                # the filter is not forwarded to the child node, and the filtering is
                # achieved locally achieved locally to this LambdasConnector.
                # As a sequel, the child node may return entries that could be filtered
                # thanks to the self.reshape_entries method.
                q_child.filters = None

        entries = self.reshape_entries(
            q,
            lambdas(
                self.m_map_lambdas,
                self.child.query(q_child),
                q.attributes
            )
        )

        return self.answer(q, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

class ForEachFilter:
    """
    The :py:class:`ForEachFilter` class is useful when an
    minifold entry (i.e., a dictionary) maps a key with a
    value which is a list of sub-entries.

    It filters each sub-entries that do not match the set
    of provided filters.
    """
    def __init__(self, attribute_name: str, map_filter: dict):
        """
        Constructor.

        Args:
            attribute_name (str): The key of the minifold entry
                attribute mapped with a list of sub-entries.
            map_filter (dict): A ``dict(sub_key : filter)``
                where `̀`sub_key`` corresponds to an (optional) sub-entry key
                and ``filter`` is a ``callable(object) -> bool`` returning
                ``True`` if and only if the value mapped to `̀`sub_key``
                indicates that the sub-entry must be filtered,
                ``False`` otherwise.
        """
        self.attribute_name = attribute_name
        self.map_filter = map_filter

    def match_filters(self, sub_entry: dict) -> bool:
        """
        Checks whether a sub-entry is filtered according to this
        :py:class:`ForEachFilter` instance.

        Args:
            sub_entry (dict): The sub-entry to be checked.

        Returns:
            ``True`` if ``sub_entry`` is not filtered,
            ``False`` otherwise.
        """
        for attribute, keep in self.map_filter.items():
            if attribute in sub_entry.keys():
                if not keep(sub_entry[attribute]):
                    return False
        return True

    def __call__(self, entry: dict) -> bool:
        """
        Functor method, filtering each sub-entry filtered
        by this :py:class:`ForEachFilter`.

        Args:
            entry (dict): The minifold entry.

        Returns:
            ``True`` if successful.
        """
        sub_entries = entry.get(self.attribute_name)

        if not sub_entries:
            return True

        entry[self.attribute_name] = [
            sub_entry
            for sub_entry in sub_entries
            if self.match_filters(sub_entry)
        ]

        return True

def for_each_sub_entry(entry, attribute: str, map_lambda: dict) -> dict:
    """
    Applies a set of lambda functions to the sub-entries carried by
    a minifold entry (key, value) pair.

    Args:
        attribute (str): The key of ``entry`` carrying the sub-entries
            to be processed.
        map_lambda (dict): A ``dict(sub_key : transform)`` where:
            ``sub_key`` is an (optional) key involved in a sub-entry;
            ``transform`` is a function transforming the value mapped
            to ``sub_key`` to its new value.

    Returns:
        The modified entry.
    """
    entry[attribute] = [
        {
            k: map_lambda[k](v) if k in map_lambda.keys() else v
            for (k, v) in sub_entry.items()
        }
        for sub_entry in entry[attribute]
    ]
    return entry



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold


class Singleton(type):
    """
    The :py:class:`Singleton` allows to define singleton classes, i.e.
    classes that can be instantiated at most once.

    >>> class MyClass(metaclass=Singleton): pass
    >>> x = MyClass()
    >>> y = MyClass()
    >>> x is y
    True

    Based on this `thread <https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python>`__.
    """
    s_instances = dict()

    def __call__(cls, *args, **kwargs):
        """
        Functor method.
        """
        if cls not in cls.s_instances:
            cls.s_instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.s_instances[cls]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

from copy import deepcopy

ACTION_CREATE = 0  # For INSERT ... queries
ACTION_READ = 1    # For SELECT ... queries
ACTION_UPDATE = 2  # For UPDATE ... queries
ACTION_DELETE = 3  # For DELETE ... queries

SORT_ASC = True    # For ... SORT BY ... ASC queries
SORT_DESC = False  # For ... SORT BY ... DESC queries

def action_to_str(action: int) -> str:
    """
    Converts an :py:class:`Query` action to its corresponding string.

    Args:
        action (int): A value in
            :py:data:`ACTION_CREATE`,
            :py:data:`ACTION_READ`,
            :py:data:`ACTION_UPDATE`,
            :py:data:`ACTION_DELETE`.

    Returns:
        The corresponding string.
    """
    if action == ACTION_CREATE:
        return "INSERT"
    elif action == ACTION_READ:
        return "SELECT"
    elif action == ACTION_UPDATE:
        return "UPDATE"
    elif action == ACTION_DELETE:
        return "DELETE"
    else:
        raise RuntimeError("action_to_str: invalid action %s" % action)

class Query:
    def __init__(
        self,
        action = ACTION_READ,
        object: str = "",
        attributes: list = None,
        filters: object = None,
        offset: int = None,
        limit: int = None,
        sort_by: dict = None
    ):
        """
        Constructor.

        Args:
            action (int): A value in
                :py:data:`ACTION_CREATE` (in SQL, INSERT queries),
                :py:data:`ACTION_READ` (in SQL, SELECT queries),
                :py:data:`ACTION_UPDATE` (in SQL, UPDATE queries),
                :py:data:`ACTION_DELETE` or ``None`` (in SQL, DELETE queries).
                Passing ``None`` is equivalent to :py:data:`ACTION_CREATE`.
                Defaults to ``None``.
            object (str): The queried object collection (some minifold gateways
                may host several entries collections, identified by a name).
                Defaults to ``""``.
            attributes (list): A list of attributes. In SQL, this corresponds
                to the attributes pass to `SELECT`.
            filters (object): A minifold filter. In SQL, this corresponds
                to the WHERE clause. See also
                the :py:class:`BinaryPredicate`a and the :py:class:`SearchFilter`
                classes.
            offset (int): A positive integer or ``None`` if not needed.
                In SQL this corresponds to the OFFSET statement.
            limit (int): A positive integer or ``None`` if not needed.
                In SQL this corresponds to the LIMIT statement.
            sort_by (dict): A dictionary characterizing how to sort the results.
                It maps each attributes to be sorted with the corresponding
                sorting order (:py:data:`SORT_ASC` or :py:data:`SORT_DESC`).
                In SQL this corresponds to the SORT BY statement.
        """
        self.m_action = action
        self.m_object = object
        self.m_attributes = list(attributes) if attributes else list()
        self.m_filters = filters

        # NOTE/TODO: the following specifiers are not supported by all the connectors
        self.m_offset = offset
        self.m_limit = limit
        self.m_sort_by = sort_by if sort_by else dict()

    def copy(self):
        """
        Copies this :py:class:`Query` instance.

        Returns:
            The copied :py:class:`Query` instance.
        """
        return deepcopy(self)

    @property
    def action(self) -> int:
        return self.m_action

    @action.setter
    def action(self, value: int):
        self.m_action = value

    @property
    def attributes(self) -> list:
        return self.m_attributes

    @attributes.setter
    def attributes(self, value: list):
        self.m_attributes = value

    @property
    def filters(self):
        return self.m_filters

    @filters.setter
    def filters(self, value):
        self.m_filters = value

    @property
    def limit(self) -> int:
        return self.m_limit

    @limit.setter
    def limit(self, value: int):
        self.m_limit = value

    @property
    def offset(self) -> int:
        return self.m_offset

    @offset.setter
    def offset(self, value: int):
        self.m_offset = value

    @property
    def sort_by(self) -> list:
        return self.m_sort_by

    @sort_by.setter
    def sort_by(self, value: dict):
        self.m_sort_by = value

    @property
    def object(self) -> str:
        return self.m_object

    def __str__(self) -> str:
        """
        Converts this :py:class:`Query` instance to its corresponding string representation.

        Returns:
            The corresponding string representation.
        """
        return "%(action)s%(attributes)s%(object)s%(filters)s%(limit)s%(offset)s%(sort_by)s" % {
            "action": action_to_str(self.action),
            "attributes": ", ".join(
                    sorted(set(self.attributes))
                ) if len(self.attributes) > 0 else " *",
            "object": " FROM %s" % self.object if self.object else "",
            "filters": " WHERE %s" % self.filters if self.filters else "",
            "limit": " LIMIT %s" % self.limit if self.limit else "",
            "offset": " OFFSET %s" % self.offset if self.offset  else "",
            "sort_by": " SORT BY %s" % ", ".join(
                    [
                        "%s %s" % (
                            attribute,
                            "ASC" if sort_asc == SORT_ASC else "DESC"
                        ) for attribute, sort_asc in self.sort_by.items()
                    ]
                ) if self.sort_by else "",
        }
